@@ -1,43 +1,54 @@
 package aQute.junit;
 
-import java.io.*;
-import java.net.*;
-import java.text.*;
-import java.util.*;
+import java.io.PrintWriter;
+import java.io.StringWriter;
+import java.io.Writer;
+import java.net.InetAddress;
+import java.net.URL;
+import java.time.Instant;
+import java.time.ZoneId;
+import java.time.format.DateTimeFormatter;
+import java.util.List;
+import java.util.Locale;
+import java.util.StringTokenizer;
+import java.util.regex.Matcher;
+import java.util.regex.Pattern;
 
-import junit.framework.*;
+import org.junit.runner.Describable;
+import org.junit.runner.Description;
+import org.osgi.framework.Bundle;
 
-import org.osgi.framework.*;
+import junit.framework.AssertionFailedError;
+import junit.framework.Test;
 
 public class JunitXmlReport implements TestReporter {
-	Tag				testsuite	= new Tag("testsuite");
-	Tag				testcase;
-	static String	hostname;
-	DateFormat		df			= new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
-	long			startTime;
-	long			testStartTime;
-	int				tests		= 0;
-	int				failures	= 0;
-	int				errors		= 0;
-	int				skipped		= 0;
-	PrintWriter		out;
-	boolean			finished;
-	boolean			progress;
-	Bundle			bundle;
-	BasicTestReport	basic;
+	private static final DateTimeFormatter	DATE_TIME_FORMATTER	= DateTimeFormatter.ISO_LOCAL_DATE_TIME
+		.withLocale(Locale.ROOT)
+		.withZone(ZoneId.systemDefault());
+	private final Tag						testsuite	= new Tag("testsuite");
+	private Tag								testcase;
+	private static String					hostname;
+	private long							startTime;
+	private long							testStartTime;
+	private final PrintWriter				out;
+	private boolean							finished;
+	private final BasicTestReport			basic;
 
 	public JunitXmlReport(Writer report, Bundle bundle, BasicTestReport basic) throws Exception {
-		if (hostname == null)
-			hostname = InetAddress.getLocalHost().getHostName();
+		try {
+			if (hostname == null)
+				hostname = InetAddress.getLocalHost()
+					.getHostName();
+		} catch (Exception e) {
+			hostname = e.getMessage();
+		}
 		out = new PrintWriter(report);
-		this.bundle = bundle;
 		this.basic = basic;
 	}
 
-	public void setProgress(boolean progress) {
-		this.progress = progress;
-	}
+	public void setProgress(boolean progress) {}
 
+	@Override
 	public void setup(Bundle fw, Bundle targetBundle) {
 		startTime = System.currentTimeMillis();
 
@@ -49,38 +60,27 @@ public class JunitXmlReport implements TestReporter {
 			testsuite.addAttribute("name", "test.run");
 
 		}
-		testsuite.addAttribute("timestamp", df.format(new Date()));
+		String timestamp = DATE_TIME_FORMATTER.format(Instant.now());
+		testsuite.addAttribute("timestamp", timestamp);
 		testsuite.addAttribute("framework", fw);
 		testsuite.addAttribute("framework-version", fw.getVersion());
 
 		Tag properties = new Tag("properties");
 		testsuite.addContent(properties);
 
-		for (Map.Entry<Object,Object> entry : System.getProperties().entrySet()) {
-			Tag property = new Tag(properties, "property");
-			property.addAttribute("name", entry.getKey());
-			property.addAttribute("value", entry.getValue());
+		for (String prop : System.getProperties()
+			.stringPropertyNames()) {
+			String value = System.getProperty(prop);
+			if (value != null) {
+				Tag property = new Tag(properties, "property");
+				property.addAttribute("name", prop);
+				property.addAttribute("value", value);
+			}
 		}
 
-		Tag bundles = new Tag(testsuite, "bundles");
-		Bundle bs[] = fw.getBundleContext().getBundles();
-
-		for (int i = 0; i < bs.length; i++) {
-			Tag bundle = new Tag("bundle");
-			bundle.addAttribute("location", bs[i].getLocation());
-			bundle.addAttribute("modified", df.format(new Date(bs[i].getLastModified())));
-			bundle.addAttribute("state", bs[i].getState());
-			bundle.addAttribute("id", bs[i].getBundleId() + "");
-			bundle.addAttribute("bsn", bs[i].getSymbolicName());
-			bundle.addAttribute("version", bs[i].getVersion());
-
-			if (bs[i].equals(targetBundle))
-				bundle.addAttribute("target", "true");
-
-			bundles.addContent(bundle);
-		}
 		if (targetBundle != null) {
-			String header = (String) targetBundle.getHeaders().get("Bnd-AddXMLToTest");
+			String header = targetBundle.getHeaders()
+				.get(aQute.bnd.osgi.Constants.BND_ADDXMLTOTEST);
 			if (header != null) {
 				StringTokenizer st = new StringTokenizer(header, " ,");
 
@@ -89,40 +89,32 @@ public class JunitXmlReport implements TestReporter {
 					URL url = targetBundle.getEntry(resource);
 
 					if (url != null) {
-						String name = url.getFile();
-						int n = name.lastIndexOf('/');
-						if (n < 0)
-							n = 0;
-						else
-							n = n + 1;
-
-						if (name.endsWith(".xml"))
-							name = name.substring(n, name.length() - 4);
-						else
-							name = name.substring(n, name.length()).replace('.', '_');
-
 						testsuite.addContent(url);
-
 					} else {
 						Tag addxml = new Tag(testsuite, "error");
-						addxml.addAttribute("reason", "no such resource: " + resource);
+						addxml.addAttribute("message", "no such resource: " + resource);
 					}
 				}
 			}
 		}
 	}
 
+	@Override
 	public void begin(List<Test> classNames, int realcount) {}
 
+	@Override
 	public void end() {
 		if (!finished) {
 			finished = true;
-			testsuite.addAttribute("tests", tests);
-			testsuite.addAttribute("failures", failures);
-			testsuite.addAttribute("errors", errors);
-			testsuite.addAttribute("skipped", skipped);
+			testsuite.addAttribute("tests", basic.getTestResult()
+				.runCount());
+			testsuite.addAttribute("failures", basic.getTestResult()
+				.failureCount());
+			testsuite.addAttribute("errors", basic.getTestResult()
+				.errorCount());
 			testsuite.addAttribute("time", getFraction(System.currentTimeMillis() - startTime, 1000));
-			testsuite.addAttribute("timestamp", df.format(new Date()));
+			String timestamp = DATE_TIME_FORMATTER.format(Instant.now());
+			testsuite.addAttribute("timestamp", timestamp);
 			testsuite.print(0, out);
 			out.close();
 		}
@@ -134,12 +126,31 @@ public class JunitXmlReport implements TestReporter {
 
 	// <testcase classname="test.AnnotationsTest" name="testComponentReader"
 	// time="0.045" />
+	private final static Pattern NAMEANDCLASS = Pattern.compile("(.*)\\((.*)\\)");
+
+	@Override
 	public void startTest(Test test) {
-		testcase = new Tag("testcase");
-		testsuite.addContent(testcase);
-		testcase.addAttribute("classname", test.getClass().getName());
 		String nameAndClass = test.toString();
 		String name = nameAndClass;
+		String clazz = test.getClass()
+			.getName();
+
+		if (test instanceof Describable) {
+			Description description = ((Describable) test).getDescription();
+			clazz = description.getClassName();
+			if (description.getMethodName() != null)
+				name = description.getMethodName();
+		} else {
+			Matcher m = NAMEANDCLASS.matcher(nameAndClass);
+			if (m.matches()) {
+				name = m.group(1);
+				clazz = m.group(2);
+			}
+		}
+
+		testcase = new Tag("testcase");
+		testsuite.addContent(testcase);
+		testcase.addAttribute("classname", clazz);
 
 		int n = nameAndClass.indexOf('(');
 		if (n > 0 && nameAndClass.endsWith(")")) {
@@ -160,17 +171,22 @@ public class JunitXmlReport implements TestReporter {
 	// </error>
 	// </testcase>
 
+	@Override
 	public void addError(Test test, Throwable t) {
 		Tag error = new Tag("error");
 		error.setCDATA();
-		error.addAttribute("type", t.getClass().getName());
+		error.addAttribute("type", t.getClass()
+			.getName());
+		String message = t.getMessage();
+		if (message != null) {
+			error.addAttribute("message", message);
+		}
 		error.addContent(getTrace(t));
 		if (testcase == null)
 			testsuite.addContent(error);
 		else
 			testcase.addContent(error);
 		progress(" e");
-		errors++;
 	}
 
 	private void progress(@SuppressWarnings("unused") String s) {}
@@ -178,14 +194,10 @@ public class JunitXmlReport implements TestReporter {
 	private String getTrace(Throwable t) {
 		StringWriter sw = new StringWriter();
 		PrintWriter pw = new PrintWriter(sw);
-		pw.println(t.toString());
-
-		StackTraceElement ste[] = t.getStackTrace();
-		for (int i = 0; i < ste.length; i++) {
-			pw.println("at " + ste[i].toString().trim());
-		}
+		t.printStackTrace(pw);
 		pw.close();
-		return sw.toString();
+		return sw.toString()
+			.replace('\t', ' ');
 	}
 
 	// <testcase classname="test.AnalyzerTest" name="testFindClass"
@@ -196,37 +208,44 @@ public class JunitXmlReport implements TestReporter {
 	// </failure>
 	// <testcase>
 	//
+	@Override
 	public void addFailure(Test test, AssertionFailedError t) {
 		Tag failure = new Tag("failure");
 		failure.setCDATA();
-		failure.addAttribute("type", t.getClass().getName());
+		failure.addAttribute("type", t.getClass()
+			.getName());
+		String message = t.getMessage();
+		if (message != null) {
+			failure.addAttribute("message", message);
+		}
 		failure.addContent(getTrace(t));
 		testcase.addContent(failure);
 		progress(" f");
-		failures++;
 	}
 
+	@Override
 	public void endTest(Test test) {
 		String[] outs = basic.getCaptured();
 		if (outs[0] != null) {
 			Tag sysout = new Tag(testcase, "system-out");
+			sysout.setCDATA();
 			sysout.addContent(outs[0]);
 		}
 
 		if (outs[1] != null) {
-			Tag sysout = new Tag(testcase, "system-err");
-			sysout.addContent(outs[1]);
+			Tag syserr = new Tag(testcase, "system-err");
+			syserr.setCDATA();
+			syserr.addContent(outs[1]);
 		}
 
 		testcase.addAttribute("time", getFraction(System.currentTimeMillis() - testStartTime, 1000));
-		tests++;
-		tests++;
 	}
 
 	public void close() {
 		end();
 	}
 
+	@Override
 	public void aborted() {
 		testsuite.addAttribute("aborted", "true");
 		close();
@@ -236,4 +255,4 @@ public class JunitXmlReport implements TestReporter {
 		testsuite.addContent(tag);
 	}
 
-}
\ No newline at end of file
+}
