@@ -1,15 +1,18 @@
 package com.example.demo;
 
-import org.osgi.framework.*;
+import org.osgi.framework.BundleActivator;
+import org.osgi.framework.BundleContext;
 
 public class Activator implements BundleActivator {
 
+	@Override
 	public void start(BundleContext arg0) throws Exception {
-		System.out.println("Hello World");
+		System.err.println("Hello World");
 	}
 
+	@Override
 	public void stop(BundleContext arg0) throws Exception {
-		System.out.println("Goodbye World");
+		System.err.println("Goodbye World");
 	}
 
 }
